package edu.columbia.xcap.servlets;

/** HttpServlet imports */
import javax.servlet.ServletConfig; 
import javax.servlet.http.HttpServlet;
import java.io.BufferedReader;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.ServletOutputStream;

/** 
 * Exceptions 
 *  -> ServletException
 */
import java.io.IOException;
import javax.servlet.ServletException;
import edu.columbia.xcap.utils.XCAPException;

/** Document Store */
import edu.columbia.xcap.store.DocumentHandler;

/** FileStore specializaion */
import edu.columbia.xcap.store.FileDocumentHandler;

/** Application usage specialization */
import edu.columbia.xcap.appusage.AppUsage;
import edu.columbia.xcap.appusage.AppUsageFactory;

import edu.columbia.xcap.utils.*;
import edu.columbia.xcap.error.*;

/** AppUsage Readers */
import java.io.File;
import java.io.FileReader;
import java.io.BufferedReader;

/** AppUsage Loaders */
import java.lang.Class;
import java.lang.ClassLoader;
import java.lang.reflect.Field;


/**
 * Main servlet class that delegates responsiblities
 */
public class XCAPServer extends HttpServlet
{
  /**
   * ------------------80 chars marking-----------------------------------------
   * Method: init
   *         Method called by Servlet Container at the initialization to
   *         store global variables to be instantiated only once. The
   *         Servlet should performe one-time setup procedures in this
   *         method and store the ServletConfig object so that it can be
   *         retrieved later by calling the Servlet's getServletConfig()
   *         method.
   * @param  ServletConfig specifies the Servlet Configuration
   * @throws ServletException
   */
  public void init(ServletConfig config) throws ServletException 
  {
    super.init(config);

    /** Create application usage and register it */
    File file = new File(SchemaConstants.CLASS_LOADER);
    if(file.exists() == false)
    {
      throw new ServletException("Could not find class loader");
    }
    else
    {
      try
      {
        /** Read application usages to be started up */
        BufferedReader bufferedFileReader = new BufferedReader(
            new FileReader(file));
        String appUsageStr = null;
        while((appUsageStr = bufferedFileReader.readLine()) != null)
        {
          /** Load the application usages */
          System.out.println("appUsageStr: " + appUsageStr);
          Class appUsageClass = Class.forName(appUsageStr);
          AppUsage usage = (AppUsage)appUsageClass.newInstance();
          AppUsageFactory.INSTANCE.registerApplicationUsage(usage);
        }
      } /** Catch specific exceptions for code clarity */
      catch(java.io.FileNotFoundException exception)
      {
        throw new ServletException(exception.getMessage());
      }
      catch(java.io.IOException exception)
      {
        throw new ServletException(exception.getMessage());
      }
      catch(java.lang.ClassNotFoundException exception)
      {
        throw new ServletException(exception.getMessage());
      }
      catch(java.lang.NoSuchFieldException exception)
      {
        throw new ServletException(exception.getMessage());
      }
      catch(java.lang.IllegalAccessException exception)
      {
        throw new ServletException(exception.getMessage());
      }
      catch(java.lang.InstantiationException exception)
      {
        throw new ServletException(exception.getMessage());
      }
    }

    try
    {
      /** Start the change notifier, XCAP-DIFF yet to be supported */
      //NotifyChange notifyChange = 
      //  NotifyChangeMgr.getInstance().getNotifyChange();
      //notifyChange.start();
      /** Send initial data to the PA server configured */
      GetInitialData getInitialData = new GetInitialData();
      getInitialData.start();
    }
    catch(XCAPException exception)
    {
      throw new ServletException(exception.getMessage());
    }
  }; // init
  

  /**
   * Method: destroy
   *         Called by garbage collector at destruction time
   * @todo   [Anurag]
   *         There may still be threads that execute the service method when
   *         destroy is called, so destroy has to be thread-safe. All
   *         resources which were allocated in init should be released in
   *         destroy. This method is guaranteed to be called only once
   *         during the Servlet's lifecycle.
   */
  public void destroy() 
  {
    /* No state variables to clear */
  }; // destroy


  /**
   * Method: doGet
   *         Called by the server (via the service method) to allow a
   *         servlet to handle a GET request. 
   *         http://java.sun.com/j2ee/sdk_1.3/techdocs/api/javax/servlet/http/
   *             HttpServlet.html#doGet(javax.servlet.http.HttpServletRequest,
   *             javax.servlet.http.HttpServletResponse)
   * @param  req HttpServletRequest
   * @param  resp HttpServletResponse
   * @throws ServletException
   * @throws IOException
   */
  protected void doGet(HttpServletRequest req, HttpServletResponse resp) 
    throws ServletException, IOException
  {
    System.out.println("-->doGet");
    /** Disable caching -- polite one */
    disableCache(resp);
    try
    {
      /** Get the XCAP URI info */
      String docNode = req.getPathInfo();
      String contextPath = req.getContextPath();
      String pathTranslated = req.getPathTranslated();
      String query = req.getQueryString();

      /** Set query to "" if not specfied in XCAP URI */
      if(query == null)
      {
        System.out.println("NULL QUERY");
        query = new String("");
      }

      /** document Selector not specified */
      if(docNode == null) 
      {
        throw new XCAPException(HttpServletResponse.SC_NOT_FOUND, 
            new NullError(SchemaConstants.NULLSCHEMA, "XCAP expression not specified"));
      }

      /** 
       * Initialize the Document Handler through FileHandler specialization
       */
      DocumentHandler docHandler = new FileDocumentHandler(docNode, query);

      /** Check the HttpScemantics of the XCAP Request */
      HttpSemantics httpSemantics = new HttpSemantics();
      if(httpSemantics.conditionalSemantics(req, resp, docHandler) == true)
      {
        /** Process get request */
        docHandler.getDocument(req, resp);
      }
      else /** HttpSemantics check failed -- mostly ETag related */
      {
        throw new XCAPException(HttpServletResponse.SC_INTERNAL_SERVER_ERROR,
            new NullError(SchemaConstants.NULLSCHEMA, 
              "Conditional handling flawed"));
      }
    }
    catch(XCAPException e)
    {
      /** Get and send the error codes */
      /** 
       * @NOTE IMPORTANT:
       * Remove repeatition of code down below 
       */
      if(e.getErrorCode().getError())
      {
        resp.setContentType(e.getErrorCode().getMimeType());
        resp.setStatus(e.getCode());
        java.io.PrintWriter pWriter = resp.getWriter();
        pWriter.print(e.getErrorCode().serialize());
      }
      else
      {
        resp.setContentType(e.getErrorCode().getMimeType());
        resp.setStatus(e.getCode());
        java.io.PrintWriter pWriter = resp.getWriter();
        pWriter.print(e.getErrorCode().getReason());
      }
    }; // catch
  }; // doGet


  /**
   * Method: doPut
   *         Called by the server (via the service method) to allow a
   *         servlet to handle a PUT request. 
   *         http://java.sun.com/j2ee/sdk_1.3/techdocs/api/javax/servlet/http/
   *             HttpServlet.html#doGet(javax.servlet.http.HttpServletRequest,
   *             javax.servlet.http.HttpServletResponse)
   * @param  req the HttpServletRequest
   * @param  resp the HttpServletResponse
   * @throws ServletException
   * @throws IOException
   */
  protected void doPut(HttpServletRequest req, HttpServletResponse resp) 
    throws ServletException, IOException 
  {
    System.out.println("-->doPut");
    disableCache(resp);
    try
    {
      /** Get the XCAP URI info */
      /**
       * @NOTE IMPORTANT:
       * Strings are non-mutable references in java, maybe we could make
       * this global for this servelt thread to avoid this code repeatation
       * --> Use call back registered handler to simulate function pointer
       *  and get the needless repeation out.
       * --> It is ok for immediate use but not for future use  
       */ 
      String docNode = req.getPathInfo();
      String contextPath = req.getContextPath();
      String pathTranslated = req.getPathTranslated();
      String query = req.getQueryString();

      /** document Selector not specified */
      if(docNode == null) 
      {
        throw new XCAPException(HttpServletResponse.SC_NOT_FOUND, 
            new NullError(SchemaConstants.NULLSCHEMA, 
              "XCAP expression not specified"));
      }

      /** Set query to "" if not specfied in XCAP URI */
      if(query == null)
      {
        System.out.println("NULL QUERY");
        query = new String("");
      }

      /** 
       * Initialize the Document Handler through FileHandler specialization
       */
      DocumentHandler docHandler = new FileDocumentHandler(docNode, query);

      /** Check the HttpScemantics of the XCAP Request */
      HttpSemantics httpSemantics = new HttpSemantics();
      if(httpSemantics.conditionalSemantics(req, resp, docHandler) == true)
      {
        System.out.println("+++++GOING FOR NEW PUT+++++");
        /** Process PUT request */
        boolean newCreation = docHandler.putDocument(req, resp);

        /** Set response message as per new creation or update */
        if(newCreation)
        {
          resp.setStatus(HttpServletResponse.SC_CREATED);
        }
        else
        {
           resp.setStatus(HttpServletResponse.SC_OK);
        }
      }
      else
      {
        throw new XCAPException(HttpServletResponse.SC_INTERNAL_SERVER_ERROR,
            new NullError(SchemaConstants.NULLSCHEMA, 
              "Conditional handling flawed"));
      }
    }
    catch(XCAPException e)
    {
      /** Get and send the error codes */
      if(e.getErrorCode().getError())
      {
        //resp.sendError(e.getCode(), e.getErrorCode().serialize());
        resp.setContentType(e.getErrorCode().getMimeType());
        resp.setStatus(e.getCode());
        java.io.PrintWriter pWriter = resp.getWriter();
        pWriter.print(e.getErrorCode().serialize());
      }
      else
      {
        resp.setContentType(e.getErrorCode().getMimeType());
        resp.setStatus(e.getCode());
        java.io.PrintWriter pWriter = resp.getWriter();
        pWriter.print(e.getErrorCode().getReason());
      }
    }; // catch

  }; // doPut

  /**
   * Method: doPost
   *         Called by the server (via the service method) to allow a
   *         servlet to handle a POST request. 
   *         http://java.sun.com/j2ee/sdk_1.3/techdocs/api/javax/servlet/http/
   *             HttpServlet.html#doGet(javax.servlet.http.HttpServletRequest,
   *             javax.servlet.http.HttpServletResponse)
   * @param  req the HttpServletRequest
   * @param  resp the HttpServletResponse
   * @throws ServletException
   * @throws IOException
   */
  protected void doPost(HttpServletRequest req, HttpServletResponse resp) 
    throws ServletException, IOException 
  {
     resp.sendError(HttpServletResponse.SC_METHOD_NOT_ALLOWED, 
         "POST method not allowed for XCAP");
  }; // doPost


  /** 
   * Method: doDelete
   *         Called by the server (via the service method) to allow a servlet 
   *         to handle a DELETE request.
   *         http://java.sun.com/j2ee/sdk_1.3/techdocs/api/javax/servlet/http/
   *           HttpServlet.html#doGet(javax.servlet.http.HttpServletRequest,
   *             javax.servlet.http.HttpServletResponse)
   * @param request servlet request
   * @param response servlet response
   * @throws ServletException
   * @throws IOException
   */
  protected void doDelete(HttpServletRequest req, HttpServletResponse resp)
    throws ServletException, IOException 
  {
    /** 
     * @NOTE IMPORTANT
     * The same code as above, please use handlers.
     * ServiceHandlers have already been implemented
     */
    try
    {
      disableCache(resp);
      String docNode = req.getPathInfo();
      String query = req.getQueryString();

      if(docNode == null) 
      {
        throw new XCAPException(HttpServletResponse.SC_NOT_FOUND, 
            new NullError(SchemaConstants.NULLSCHEMA, 
              "XCAP expression not specified"));
      }

      if(query == null)
      {
        System.out.println("NULL QUERY");
        query = new String("");
      }
      
      DocumentHandler docHandler = new FileDocumentHandler(docNode, query);

      HttpSemantics httpSemantics = new HttpSemantics();

      if(httpSemantics.conditionalSemantics(req, resp, docHandler) == true)
      {
        docHandler.deleteDocument();
        resp.setStatus(HttpServletResponse.SC_OK);
      }
      else
      {
        throw new XCAPException(HttpServletResponse.SC_INTERNAL_SERVER_ERROR,
            new NullError(SchemaConstants.NULLSCHEMA, 
              "Conditional handling flawed"));
      }
    }
    catch(XCAPException e)
    {
      /** Get and send the error codes */
      if(e.getErrorCode().getError())
      {
        resp.setContentType(e.getErrorCode().getMimeType());
        resp.setStatus(e.getCode());
        java.io.PrintWriter pWriter = resp.getWriter();
        pWriter.print(e.getErrorCode().serialize());
      }
      else
      {
        resp.setContentType(e.getErrorCode().getMimeType());
        resp.setStatus(e.getCode());
        java.io.PrintWriter pWriter = resp.getWriter();
        pWriter.print(e.getErrorCode().getReason());
      }
    }; // catch
  }; // doDelete


  /**
   * Method: disableCache
   *         Method to disable browser cache
   */
  void disableCache(HttpServletResponse resp) 
  {
    resp.setHeader("Pragma", "No-cache");
    resp.setHeader("Cache-Control", "no-cache");
    resp.setDateHeader("Expires", 0);
  }; // disableCache


}; // class XCAPServer
