/**
 * "MRUCache.java (c) Fall 2005 Columbia University"
 */

/** MRUCache package */
package edu.columbia.xcap.utils;


/** Hashmap as my cache */
import java.util.Map;
import java.util.HashMap;

/** MRU entries*/
import java.util.LinkedList;
import java.util.Collections;

/** Value in the map */
import org.w3c.dom.Node;

/** 
 * Class to read a docSelector ie implementation specific updating cache 
 * -> For filesystem implementation we need to read a file
 * -> For DB implementation we need to query a DB
 * @NOTE: Will be developed further only if required
 * Simple Most recently used cache implementation
 */
import edu.columbia.xcap.store.DocumentHandler;

/** */
import edu.columbia.xcap.store.DocumentHandler;

import edu.columbia.xcap.utils.MRUObject;


/**
 * MRUCache, keep the most recently used in cache
 * Standard MRU pattern of hashmap and linkedlist
 * The LinkedList provides the queue mechanism, and the entries in the
 * LinkedList contain the key to the data in the HashMap.
 * Key   --> Document Selector
 * Value --> DOM Node
 */
public class MRUCache
{
  //----------------------------------------------------------------------------
  /**
   * Method: MRUCache
   *         Constructor to construct empty cache of the given size
   * @param  maxCacheSize specifies the max cache size
   */
  public MRUCache(int maxCacheSize, DocumentHandler docHandler)
  {
    _myCache      = (Map)Collections.synchronizedMap(new HashMap());
    _mruList      = (LinkedList)Collections.synchronizedList(new LinkedList());
    _maxCacheSize = maxCacheSize;
    _docHandler   = docHandler;
    _policy       = "copy-on-read";
  }; // MRUCache


  /**
   * Method: getNode
   *         Method to get a node from the cache
   */
  public Node getNode(String docSelector, String lastModTime)
  {
    Node node = null;
    synchronized(_myCache)
    {
      if(_myCache.containsKey(docSelector))
      {
        MRUObject mruObject = (MRUObject)_myCache.get(docSelector);
        if(mruObject.getLastModTime() == lastModTime)
        {
          node = mruObject.getNode();
        }
        else
        {
          node = _docHandler.getNewNode(lastModTime);
          _mruList.addFirst(docSelector);
          _myCache.put(docSelector, (Object)new MRUObject(lastModTime, node));
        }
      }
      else 
      {
        if(_mruList.size() >= _maxCacheSize)
        {
          _myCache.remove(_mruList.getLast());
          _mruList.removeLast();
        }
        node = _docHandler.getNewNode(lastModTime);
        _mruList.addFirst(docSelector);
        _myCache.put(docSelector, (Object)new MRUObject(lastModTime, node));
      }
    }
    return node;
  }; // getNode


  /** 
   * Method: putNode
   *         Method to put into the cache
   */
  public void putNode(String docSelector, String lastModTime, Node node)
  {
    if(_policy == "copy-on-write")
    {
      synchronized(_myCache)
      {
        if(_mruList.size() >= _maxCacheSize)
        {
          _myCache.remove(_mruList.getLast());
          _mruList.removeLast();
        }
        _mruList.addFirst(docSelector);
        _myCache.put(docSelector, (Object)new MRUObject(lastModTime, node)); 
      }
    }
  }; // putNode
  

  /**
   * Method: deleteNode
   *         Method to delete a node from the cache
   */
  public void deleteNode(String docSelector)
  {
    synchronized(_myCache)
    {
      _myCache.remove(docSelector);
      _mruList.remove(docSelector);
    }
  }; // deleteNode


  /** Map as my cache */
  private Map _myCache;

  /** MRU List */
  private LinkedList _mruList;

  /** Cache capacity */
  private int _maxCacheSize;

  /** Implementation specific method to update a cache */
  private DocumentHandler _docHandler;

  /** 
   * Policy of caching 
   * copy-on-read : lazy update where it is bought to cache 
   *                only when get fails.
   * copy-on-write: get to cache immediately on write              
   */
  private String _policy;

}; // class MRUCache
